//
//  iPhonePlatformInterface.cpp
//  scummvm
//
//  Created by Omer Gilad on 9/14/15.
//
//

#include "iPhonePlatformInterface.h"
#include "enhanced/TextureAllocator.h"
#include "enhanced/GameOverlayManager.h"

#include "enhanced/GameSettingsManager.h"

#include "OpenGLESDebug.h"
#include "osys_main.h"

#import <sys/utsname.h>
//#import <Foundation/NSProcessInfo.h>

// Just for logs
#include "common/textconsole.h"

#define GL_TEXTURE_STATIC 0
#define GL_TEXTURE_DYNAMIC 1

const char *s_originalPowerMachines[] = { // Original Graphics as default
    "i386",
    "x86_32",
    "x86_64",
    "iPhone3,1",
    "iPhone3,2",
    "iPhone3,3",
    "iPad1,1",
    "iPad1,2",
    0
};

const char *s_lowPowerMachines[] = { // Medium Graphics as default
    "iPod4,1",
    "iPod5,1",
    "iPad2,1",
    "iPad2,1",
    "iPad2,2",
    "iPad2,3",
    "iPad2,4",
    "iPad2,5",
    "iPad2,6",
    "iPad2,7",
#if defined(TARGET_IHNM) || defined(TARGET_T7G) || defined(TARGET_TONYT)
    "iPad3,1",
    "iPad3,2",
    "iPad3,3",
    "iPad3,4",
    "iPad3,5",
    "iPad3,6",
    "iPad4,1",
    "iPad4,2",
    "iPad4,3",
    "iPad4,4",
    "iPad4,5",
    "iPad4,6",
    "iPad4,7",
    "iPad4,8",
    "iPad4,9",
#endif
    "iPhone4,1",
    "iPhone4,2",
    "iPhone4,3",
    0
};

const char *s_iPhoneXMachines[] = { // iPhone X identifiers
    "iPhone10,3",
    "iPhone10,6",
    "iPhone11,2",
    "iPhone11,4",
    "iPhone11,6",
    "iPhone11,8",
    0
};

const char *s_NoMT32Machines[] = { // MT-32 music doesn't work well on the following devices
    0
};

IPhonePlatformInterface::IPhonePlatformInterface() : mOverlayGlTextures(NULL), mUiInterface(NULL)
{
    mEmptyClearingBuffer = new uint16[GL_DYNAMIC_TEXTURE_WIDTH
                                      * GL_DYNAMIC_TEXTURE_HEIGHT];
    memset(mEmptyClearingBuffer, 0,
           GL_DYNAMIC_TEXTURE_WIDTH * GL_DYNAMIC_TEXTURE_HEIGHT);
}

IPhonePlatformInterface::~IPhonePlatformInterface()
{
    GLCALL(glDeleteTextures(2, mOverlayGlTextures));
    delete[] mOverlayGlTextures;
    mOverlayGlTextures = NULL;
    
    delete[] mEmptyClearingBuffer;

    delete mUiInterface;
}

Enhanced::NativeUiInterface* IPhonePlatformInterface::getNativeUiInterface()
{
    if (mUiInterface == NULL)
    {
        mUiInterface = new iPhoneNativeUiInterface;
    }
    
    return mUiInterface;
}


void IPhonePlatformInterface::drawOverlayBitmap(Enhanced::OverlayBitmap* bitmap, float x, float y,
                                                float width, float height, float alpha, Enhanced::OverlayShaderType shaderType)
{
 //   LOGD("IPhonePlatformInterface::drawOverlayBitmap %s %f %f %f %f %f", bitmap->bitmapName.c_str(), x, y, width, height, //alpha);
    
    uint16 glTextureIndex =
    bitmap->glTexture ? GL_TEXTURE_STATIC : GL_TEXTURE_DYNAMIC;
    GLuint currentTextureWidth =
    bitmap->glTexture ? GL_OVERLAY_TEXTURE_WIDTH : GL_DYNAMIC_TEXTURE_WIDTH;
    GLuint currentTextureHeight =
    bitmap->glTexture ? GL_OVERLAY_TEXTURE_HEIGHT : GL_DYNAMIC_TEXTURE_HEIGHT;
    
    GLCALL(glBindTexture(GL_TEXTURE_2D, mOverlayGlTextures[glTextureIndex]));
    
    GLfloat tex_x, tex_y, tex_width, tex_height;
    
    // Texture coordinates should be in GL tex coordinate system(0.0 - 1.0)
    if (bitmap->glTexture) {
        // If the bitmap is already in a GL texture, get the stored coordinates
        tex_x = bitmap->glTextureX / (GLfloat) currentTextureWidth;
        tex_y = bitmap->glTextureY / (GLfloat) currentTextureHeight;
        tex_width = bitmap->getWidth() / (GLfloat) currentTextureWidth;
        tex_height = bitmap->getHeight() / (GLfloat) currentTextureHeight;
    } else {
        //
        // If not, we need to write the bitmap to the dynamic GL texture
        //
        // Clear the dynamic texture to prevent edge artifacts
        GLCALL(glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, currentTextureWidth, currentTextureHeight, GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, mEmptyClearingBuffer));
        
        // Write the new bitmap
        GLCALL(glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, bitmap->getWidth(), bitmap->getHeight(), GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, bitmap->getPixels()));
        
        tex_x = 0;
        tex_y = 0;
        tex_width = bitmap->getWidth() / (GLfloat) currentTextureWidth;
        tex_height = bitmap->getHeight() / (GLfloat) currentTextureHeight;
    }
    
    // define texCoords (triangle strip) - what part of the texture to crop
    // (invert y axis)
    const GLfloat texcoords[] = {tex_x, tex_y, tex_x + tex_width, tex_y, tex_x,
        tex_y + tex_height, tex_x + tex_width, tex_y + tex_height,};
    
    GLfloat bmpDisplayWidth, bmpDisplayHeight;
    if (width <= 0) {
        // In this scenario, the bitmap dimension are specified in pixels (normal screen coordinates)
        bmpDisplayWidth = bitmap->displayWidth;
        bmpDisplayHeight = bitmap->displayHeight;
    } else {
        // Set a custom display width for the bitmap, just for now
        bmpDisplayWidth = width * (GLfloat) Enhanced::GameOverlayManager::instance()->getDisplayWidth();
        if (height <= 0) {
            bmpDisplayHeight = bmpDisplayWidth * bitmap->getRatio();
        } else {
            bmpDisplayHeight = height * (GLfloat) Enhanced::GameOverlayManager::instance()->getDisplayHeight();
        }
    }
    
    x = x * (float)Enhanced::GameOverlayManager::instance()->getDisplayWidth();
    y = y * (float)Enhanced::GameOverlayManager::instance()->getDisplayHeight();
    
 //   LOGD("IPhonePlatformInterface::drawOverlayBitmap vertices: %f %f %f %f %f %f %f %f", vertices[0], vertices[1], //vertices[2], vertices[3], vertices[4], vertices[5], vertices[6], vertices[7]);
    
  //  LOGD("IPhonePlatformInterface::drawOverlayBitmap texCoords: %f %f %f %f %f %f %f %f", texcoords[0], texcoords[1], //texcoords[2], texcoords[3], texcoords[4], texcoords[5], texcoords[6], texcoords[7]);
    
    VideoContext* videoContext = ((OSystem_IPHONE*)g_system)->getVideoContext();
    ShaderInfo* shader = nullptr;
    if (shaderType == Enhanced::OverlayShaderType::NORMAL)
    {
        shader = bitmap->drawWithScaling ? videoContext->currentShader : videoContext->defaultShader;
    }
    else if (shaderType == Enhanced::OverlayShaderType::AREA)
    {
        shader = videoContext->areaShader;
    }
        
    if (shader)
    {
        const ShaderPass &p = shader->passes[0];
        GLCALL(glUseProgram(p.program));
        GLCALL(glUniform1i(p.textureLoc, 0));
        GLCALL(glUniform1f(p.alphaFactorLoc, alpha));
        GLCALL(glUniformMatrix4fv(p.projMat, 1, GL_FALSE, videoContext->projMat));
        
        if (shaderType == Enhanced::OverlayShaderType::NORMAL)
        {
            // Define vertices for drawing to the screen
            const GLfloat vertices[] = {x, y, x + bmpDisplayWidth, y, x, y
                + bmpDisplayHeight, x + bmpDisplayWidth, y + bmpDisplayHeight};
            
            // Pass the position attributes
            GLCALL(glVertexAttribPointer(p.positionAttributeLoc, 2, GL_FLOAT, false, 0, vertices));
            
            // Pass the texture coordinate attributes
            GLCALL(glVertexAttribPointer(p.texCoordAttributeLoc, 2, GL_FLOAT, false, 0, texcoords));
        }
        else if (shaderType == Enhanced::OverlayShaderType::AREA)
        {
            //
            // WORKAROUND: Push both position data and static interpolation data to the position attribute,
            // due to weird iOS bug that prevents compiling the shader with 2 different attributes.
            //
            
            // define vertices for drawing to the screen
            const GLfloat vertices[] = {x, y, -1.0, -1.0, x + bmpDisplayWidth, y, -1.0, 1.0, x, y
                + bmpDisplayHeight, 1.0, -1.0, x + bmpDisplayWidth, y + bmpDisplayHeight, 1.0, 1.0};
            
            // Pass the position attributes
            GLCALL(glVertexAttribPointer(p.positionAttributeLoc, 4, GL_FLOAT, false, 0, vertices));
        }
        
        if (bitmap->drawWithScaling) {
            GLCALL(
                   glUniform2f(p.textureSizeLoc,
                               currentTextureWidth, currentTextureHeight));
            GLCALL(
                   glUniform2f(p.textureFractLoc,
                               1 / (GLfloat )currentTextureWidth,
                               1 / (GLfloat )currentTextureHeight));
            
            GLCALL(
                   glUniform2f(p.inputSizeLoc,
                               bitmap->getWidth(), bitmap->getHeight()));
            GLCALL(
                   glUniform2f(p.outputSizeLoc,
                               bmpDisplayWidth,
                               bmpDisplayHeight));
        }
        
        // WORKAROUND: removed due to iOS bug
        // Pass the optional static interpolation
       /* if (p.staticInterpolationLoc != -1) {
            static GLfloat staticInterpolation[] = {-1.0, -1.0, -1.0, 1.0, 1.0,
                -1.0, 1.0, 1.0};
            GLCALL(
                   glVertexAttribPointer(p.staticInterpolationLoc, 2, GL_FLOAT, false, 0, staticInterpolation));
        }*/
    }
    
    GLCALL(glDrawArrays(GL_TRIANGLE_STRIP, 0, 4));
}

void IPhonePlatformInterface::allocateBitmapInTexture(Enhanced::OverlayBitmap* bitmap)
{
    LOGD("IPhonePlatformInterface::allocateBitmapInTexture %s", bitmap->bitmapName.c_str());
    
    if (mOverlayGlTextures == NULL)
    {
        initOverlayTextures();
    }
    
    GLCALL(glBindTexture(GL_TEXTURE_2D, mOverlayGlTextures[GL_TEXTURE_STATIC]));

    GLCALL(glTexSubImage2D(GL_TEXTURE_2D, 0, bitmap->glTextureX, bitmap->glTextureY, bitmap->getWidth(), bitmap->getHeight(), GL_RGBA, GL_UNSIGNED_BYTE, bitmap->getPixels()));
}

void IPhonePlatformInterface::initOverlayTextures()
{
    LOGD("IPhonePlatformInterface::initOverlayTextures()");
    
    GLCALL(glPixelStorei(GL_UNPACK_ALIGNMENT, 2));
    
    mOverlayGlTextures = new GLuint[2];
    GLCALL(glGenTextures(2, mOverlayGlTextures));
    
    // Initialize the texture for static images
    GLCALL(glBindTexture(GL_TEXTURE_2D, mOverlayGlTextures[GL_TEXTURE_STATIC]));
    GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR));
    GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR));

    GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE));
    GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE));
    byte* combinedGlTexturePixels = new byte[(GL_OVERLAY_TEXTURE_WIDTH) * (GL_OVERLAY_TEXTURE_HEIGHT) * 4];
    
    GLCALL(glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, GL_OVERLAY_TEXTURE_WIDTH, GL_OVERLAY_TEXTURE_HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, combinedGlTexturePixels));
    delete[] combinedGlTexturePixels;
    
    // Initialize the texture for dynamic bitmap drawing
    GLCALL(glBindTexture(GL_TEXTURE_2D, mOverlayGlTextures[GL_TEXTURE_DYNAMIC]));
    GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST));
    GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST));

    
    GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE));
    GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE));
    uint16* dynamicTexturePixels = new uint16[(GL_DYNAMIC_TEXTURE_WIDTH) * (GL_DYNAMIC_TEXTURE_HEIGHT)];
    
    GLCALL(glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, GL_DYNAMIC_TEXTURE_WIDTH, GL_DYNAMIC_TEXTURE_WIDTH, 0, GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, dynamicTexturePixels));
    delete[] dynamicTexturePixels;
}

bool IPhonePlatformInterface::IsMT32SupportedDevice()
{
    struct utsname systemInfo;
    uname(&systemInfo);
    
    for (int m = 0; s_NoMT32Machines[m]; m++) {
        if (!strcmp(systemInfo.machine, s_NoMT32Machines[m]))
            return false;
    }
    
    return true;
}

bool IPhonePlatformInterface::IsiPhoneXDevice()
{
    struct utsname systemInfo;
    uname(&systemInfo);
 
    char *machine;
#if TARGET_IPHONE_SIMULATOR
    machine = getenv("SIMULATOR_MODEL_IDENTIFIER");
#else
    machine = systemInfo.machine;
#endif
    
    for (int m = 0; s_iPhoneXMachines[m]; m++) {
        if (!strcmp(machine, s_iPhoneXMachines[m]))
            return true;
    }
    
    return false;
}


Common::String IPhonePlatformInterface::getDefaultGraphicSetting()
{
    struct utsname systemInfo;
    uname(&systemInfo);
    printf("System machine: =%s",systemInfo.machine);
  
    for (int m = 0; s_originalPowerMachines[m]; m++) {
        if (!strcmp(systemInfo.machine, s_originalPowerMachines[m]))
            return SETTING_VALUE_GRAPHIC_MODE_ORIGINAL;
    }
    
    for (int m = 0; s_lowPowerMachines[m]; m++) {
        if (!strcmp(systemInfo.machine, s_lowPowerMachines[m]))
            return SETTING_VALUE_GRAPHIC_MODE_LQ;
    }
    
    return SETTING_VALUE_GRAPHIC_MODE_HQ;
}

Common::String IPhonePlatformInterface::getGameFilePath()
{
    return Common::String(".");
}



