//
//  Win32PlatformInterface.cpp
//  scummvm
//
//
//

#define NOMINMAX
#define GL_GLEXT_PROTOTYPES

#include "backends/platform/sdl/win32/Win32PlatformInterface.h"
#include "backends/graphics/opengl/debug.h"
#include "backends/graphics/opengl/opengl-graphics.h"
#include "backends/modular-backend.h"

#include "graphics/scaler/xBRZ/xbrz.h"
#include "graphics/conversion.h"

#include "enhanced/GameSettingsManager.h"
#include "enhanced/GameOverlayManager.h"
#include "enhanced/constants/GameInfo.h"
#include "enhanced/TextureAllocator.h"
#include "enhanced/constants/Constants.h"


// Just for logs
#include "common/textconsole.h"

const char *s_areaVertex =
"precision highp float;\n"
"attribute vec4 vPosition;\n"
"uniform mat4 projMat;\n"
"varying vec2 interpolator;\n"
""
"void main() {\n"
"   interpolator = vec2(vPosition.z, vPosition.w);\n"
"  gl_Position = gl_ModelViewProjectionMatrix * vec4(vPosition.x, vPosition.y, 0.0, 1.0);\n"
"}\n"
;

const char *s_areaFragment =
"precision highp float;\n"
""
"uniform float alphaFactor;\n"
"varying vec2 interpolator;\n"
""
"float MAX_ALPHA = 0.75;\n"
""
"void main() {\n"
"   float x = distance(interpolator, vec2(0.0, 0.0));\n"
"  float alpha = MAX_ALPHA - MAX_ALPHA * x;\n"
"   alpha = max(alpha, 0.0);\n"
"  gl_FragColor = vec4(0.925, 0.921, 0.843, alpha);\n"
"   gl_FragColor.a *= alphaFactor;\n"
"}\n";

Win32PlatformInterface::Win32PlatformInterface() : mUiInterface(NULL), mOverlayGlTextures(NULL)
{
	mEmptyClearingBuffer = new uint16[GL_DYNAMIC_TEXTURE_WIDTH
		* GL_DYNAMIC_TEXTURE_HEIGHT];
	memset(mEmptyClearingBuffer, 0,
		GL_DYNAMIC_TEXTURE_WIDTH * GL_DYNAMIC_TEXTURE_HEIGHT * sizeof(uint16));

	mEmptyClearingBuffer2 = new uint32[GL_DYNAMIC_TEXTURE_WIDTH
		* GL_DYNAMIC_TEXTURE_HEIGHT];
	memset(mEmptyClearingBuffer2, 0,
		GL_DYNAMIC_TEXTURE_WIDTH * GL_DYNAMIC_TEXTURE_HEIGHT * sizeof(uint32));

	mScalerOutputARGB8888 = new uint32[GL_DYNAMIC_TEXTURE_WIDTH
		* GL_DYNAMIC_TEXTURE_HEIGHT];
	memset(mScalerOutputARGB8888, 0,
		GL_DYNAMIC_TEXTURE_WIDTH * GL_DYNAMIC_TEXTURE_HEIGHT * sizeof(uint32));

	mScalerInputARGB8888 = new uint32[GL_DYNAMIC_TEXTURE_WIDTH
		* GL_DYNAMIC_TEXTURE_HEIGHT];
	memset(mScalerInputARGB8888, 0,
		GL_DYNAMIC_TEXTURE_WIDTH * GL_DYNAMIC_TEXTURE_HEIGHT * sizeof(uint32));

	mScalerOutputRGBA5551 = new uint16[GL_DYNAMIC_TEXTURE_WIDTH
		* GL_DYNAMIC_TEXTURE_HEIGHT];
	memset(mScalerOutputRGBA5551, 0,
		GL_DYNAMIC_TEXTURE_WIDTH * GL_DYNAMIC_TEXTURE_HEIGHT * sizeof(uint16));
}

Win32PlatformInterface::~Win32PlatformInterface()
{
	GLCALL(glDeleteTextures(2, mOverlayGlTextures));
	delete[] mOverlayGlTextures;
	mOverlayGlTextures = NULL;

	delete[] mEmptyClearingBuffer;
	delete[] mEmptyClearingBuffer2;
	delete[] mScalerOutputARGB8888;
	delete[] mScalerInputARGB8888;
	delete[] mScalerOutputRGBA5551;

    delete mUiInterface;

	if (GameInfo::getGameType() == GAME_TYPE_T7G)
	{
		// Clear shaders
	/*	glDeleteShader(mAreaVertex);
		glDeleteShader(mAreaFragment);
		glDeleteProgram(mAreaProgram);*/
	}
}

Enhanced::NativeUiInterface* Win32PlatformInterface::getNativeUiInterface()
{
    if (mUiInterface == NULL)
    {
        mUiInterface = new Win32NativeUiInterface();
    }
    
    return mUiInterface;
}


void Win32PlatformInterface::drawOverlayBitmap(Enhanced::OverlayBitmap* bitmap, float x, float y,
                                                float width, float height, float alpha, Enhanced::OverlayShaderType shaderType)
{
 
	uint16 glTextureIndex =
		bitmap->glTexture ? GL_TEXTURE_STATIC : GL_TEXTURE_DYNAMIC;

	if (!bitmap->glTexture && bitmap->bpp == 4)
		glTextureIndex = GL_TEXTURE_DYNAMIC2;

	GLuint currentTextureWidth =
		bitmap->glTexture ? GL_OVERLAY_TEXTURE_WIDTH : GL_DYNAMIC_TEXTURE_WIDTH;
	GLuint currentTextureHeight =
		bitmap->glTexture ? GL_OVERLAY_TEXTURE_HEIGHT : GL_DYNAMIC_TEXTURE_HEIGHT;

	GLCALL(glBindTexture(GL_TEXTURE_2D, mOverlayGlTextures[glTextureIndex]));

	GLCALL(glColor4f(1.0f, 1.0f, 1.0f, alpha));

	GLfloat tex_x, tex_y, tex_width, tex_height;

	// Texture coordinates should be in GL tex coordinate system(0.0 - 1.0)
	if (bitmap->glTexture) {
		// If the bitmap is already in a GL texture, get the stored coordinates
		tex_x = bitmap->glTextureX / (GLfloat)currentTextureWidth;
		tex_y = bitmap->glTextureY / (GLfloat)currentTextureHeight;
		tex_width = bitmap->getWidth() / (GLfloat)currentTextureWidth;
		tex_height = bitmap->getHeight() / (GLfloat)currentTextureHeight;
	}
	else {
		//
		// If not, we need to write the bitmap to the dynamic GL texture
		//

		// Clear the dynamic texture to prevent edge artifacts
		if (bitmap->bpp == 4)
		{
			GLCALL(glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, currentTextureWidth, currentTextureHeight, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, mEmptyClearingBuffer2));
		}
		else
		{	
			GLCALL(glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, currentTextureWidth, currentTextureHeight, GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, mEmptyClearingBuffer));
		}

		// Check if we need to scale with XBRZ
		if (GameSettingsManager::instance()->getStringValue(SETTING_KEY_GRAPHIC_MODE) == SETTING_VALUE_GRAPHIC_MODE_HQ
			&& bitmap->drawWithScaling)
		{
			// Determine a good scaling factor
			uint16 scaleFactor = GameOverlayManager::instance()->getDisplayWidth() / GAME_SCREEN_WIDTH;
			scaleFactor = CLIP<uint16>(scaleFactor, 2, 6);
			uint16 bitmapScaledWidth = bitmap->getWidth() * scaleFactor;
			uint16 bitmapScaledHeight = bitmap->getHeight() * scaleFactor;

			Graphics::PixelFormat formatARGB8888 = Graphics::PixelFormat(4, 8, 8, 8, 8, 16, 8, 0, 24);
			Graphics::PixelFormat formatRGBA5551 = Graphics::PixelFormat(2, 5, 5, 5, 1, 11, 6, 1, 0);

			// Convert to 32bit ARGB for scaler input
			Graphics::crossBlit((byte*)mScalerInputARGB8888, (byte*)bitmap->getPixels(), 
				bitmap->getWidth() * sizeof(uint32), bitmap->getWidth() * sizeof(uint16), bitmap->getWidth(),
				bitmap->getHeight(), formatARGB8888, formatRGBA5551);

			// Scale to the output buffer
			xbrz::scale(scaleFactor, mScalerInputARGB8888, mScalerOutputARGB8888, bitmap->getWidth(), 
				bitmap->getHeight(), xbrz::ColorFormat::ARGB);

			// Convert back to 16bit RGBA for texture
			Graphics::crossBlit((byte*)mScalerOutputRGBA5551, (byte*)mScalerOutputARGB8888, 
				bitmapScaledWidth * sizeof(uint16), bitmapScaledWidth * sizeof(uint32),
				bitmapScaledWidth, bitmapScaledHeight, formatRGBA5551,
				formatARGB8888);

			// Write the new bitmap
			GLCALL(glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, bitmapScaledWidth, bitmapScaledHeight, GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, mScalerOutputRGBA5551));

			tex_width = bitmapScaledWidth / (GLfloat)currentTextureWidth;
			tex_height = bitmapScaledHeight / (GLfloat)currentTextureHeight;
		}
		else
		{
			// Write the new bitmap
			if (bitmap->bpp == 4)
			{
				GLCALL(glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, bitmap->getWidth(), bitmap->getHeight(), GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, bitmap->getPixels()));
			}
			else
			{
				GLCALL(glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, bitmap->getWidth(), bitmap->getHeight(), GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, bitmap->getPixels()));
			}
		
			tex_width = bitmap->getWidth() / (GLfloat)currentTextureWidth;
			tex_height = bitmap->getHeight() / (GLfloat)currentTextureHeight;
		}

		tex_x = 0;
		tex_y = 0;	
	}

	// define texCoords (triangle strip) - what part of the texture to crop
	// (invert y axis)
	const GLfloat texcoords[] = { tex_x, tex_y, tex_x + tex_width, tex_y, tex_x,
		tex_y + tex_height, tex_x + tex_width, tex_y + tex_height, };
	GLCALL(glTexCoordPointer(2, GL_FLOAT, 0, texcoords));

	//
	// Define vertices
	//
	GLfloat bmpDisplayWidth, bmpDisplayHeight;
	if (width <= 0) {
		// In this scenario, the bitmap dimension are specified in pixels (normal screen coordinates)
		bmpDisplayWidth = bitmap->displayWidth;
		bmpDisplayHeight = bitmap->displayHeight;
	}
	else {
		// Set a custom display width for the bitmap, just for now
		bmpDisplayWidth = width * (GLfloat)Enhanced::GameOverlayManager::instance()->getDisplayWidth();
		if (height <= 0) {
			bmpDisplayHeight = bmpDisplayWidth * bitmap->getRatio();
		}
		else {
			bmpDisplayHeight = height * (GLfloat)Enhanced::GameOverlayManager::instance()->getDisplayHeight();
		}
	}

	x = x * (float)Enhanced::GameOverlayManager::instance()->getDisplayWidth();
	y = y * (float)Enhanced::GameOverlayManager::instance()->getDisplayHeight();

	// Adjust the coordinates to actual display area - because of aspect ratio correction
	x += mDisplayOffsetX;
	y += mDisplayOffsetY;

	/*if (shaderType == Enhanced::OverlayShaderType::AREA)
	{
		GLCALL(glUseProgram(mAreaProgram));
		GLCALL(glUniform1f(mAreaAlphaFactorUniform, alpha));
	//	GLCALL(glUniformMatrix4fv(mAreaProjMatUniform, 1, GL_FALSE, videoContext->projMat));

		// define vertices for drawing to the screen
		const GLfloat vertices[] = { x, y, -1.0, -1.0, x + bmpDisplayWidth, y, -1.0, 1.0, x, y
			+ bmpDisplayHeight, 1.0,
			-1.0, x + bmpDisplayWidth, y + bmpDisplayHeight, 1.0, 1.0 };

		// Pass the position attributes
		GLCALL(glVertexAttribPointer(mAreaPositionAttribute, 4, GL_FLOAT, false, 0, vertices));
	}
	else
	{
		GLCALL(glUseProgram(0));*/

		const GLfloat vertices[] = { x, y, x + bmpDisplayWidth, y, x, y
			+ bmpDisplayHeight, x + bmpDisplayWidth, y + bmpDisplayHeight };
		GLCALL(glVertexPointer(2, GL_FLOAT, 0, vertices));
	//}

	// Draw the texture to the screen buffer.
	GLCALL(glDrawArrays(GL_TRIANGLE_STRIP, 0, 4));

	// Reset color
	GLCALL(glColor4f(1.0f, 1.0f, 1.0f, 1.0f));
}

 void Win32PlatformInterface::clearHWTexture()
{
	GLCALL(glDeleteTextures(2, mOverlayGlTextures));
	delete[] mOverlayGlTextures;
	mOverlayGlTextures = NULL;
}


void Win32PlatformInterface::allocateBitmapInTexture(Enhanced::OverlayBitmap* bitmap)
{
	if (mOverlayGlTextures == NULL)
	{
		initOverlayTextures();
	}

	GLCALL(glBindTexture(GL_TEXTURE_2D, mOverlayGlTextures[GL_TEXTURE_STATIC]));

	GLCALL(glTexSubImage2D(GL_TEXTURE_2D, 0, bitmap->glTextureX, bitmap->glTextureY, bitmap->getWidth(), bitmap->getHeight(), GL_RGBA, GL_UNSIGNED_BYTE, bitmap->getPixels()));

}

void Win32PlatformInterface::initOverlayTextures()
{
	LOGD("Win32PlatformInterface::initOverlayTextures()");

	GLCALL(glPixelStorei(GL_UNPACK_ALIGNMENT, 2));

	mOverlayGlTextures = new GLuint[3];
	GLCALL(glGenTextures(3, mOverlayGlTextures));

	// Initialize the texture for static images
	GLCALL(glBindTexture(GL_TEXTURE_2D, mOverlayGlTextures[GL_TEXTURE_STATIC]));
	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR));
	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR));

	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE));
	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE));

	byte* combinedGlTexturePixels = new byte[GL_OVERLAY_TEXTURE_WIDTH * GL_OVERLAY_TEXTURE_HEIGHT * 4];
	memset(combinedGlTexturePixels, 0,
		GL_OVERLAY_TEXTURE_WIDTH * GL_OVERLAY_TEXTURE_HEIGHT * 4);

	GLCALL(glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, GL_OVERLAY_TEXTURE_WIDTH, GL_OVERLAY_TEXTURE_HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, combinedGlTexturePixels));
	delete[] combinedGlTexturePixels;

	// Initialize the texture for dynamic bitmap drawing
	GLCALL(glBindTexture(GL_TEXTURE_2D, mOverlayGlTextures[GL_TEXTURE_DYNAMIC]));
	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST));
	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST));

	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE));
	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE));
	uint16* dynamicTexturePixels = new uint16[GL_DYNAMIC_TEXTURE_WIDTH
		* GL_DYNAMIC_TEXTURE_HEIGHT];

	GLCALL(glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, GL_DYNAMIC_TEXTURE_WIDTH, GL_DYNAMIC_TEXTURE_WIDTH, 0, GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, dynamicTexturePixels));
	delete[] dynamicTexturePixels;

	GLCALL(glBindTexture(GL_TEXTURE_2D, mOverlayGlTextures[GL_TEXTURE_DYNAMIC2]));
	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR));
	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR));

	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE));
	GLCALL(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE));
	uint32* dynamicTexturePixels2 = new uint32[GL_DYNAMIC_TEXTURE_WIDTH
		* GL_DYNAMIC_TEXTURE_HEIGHT];

	GLCALL(glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, GL_DYNAMIC_TEXTURE_WIDTH, GL_DYNAMIC_TEXTURE_WIDTH, 0, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, dynamicTexturePixels2));
	delete[] dynamicTexturePixels2;

	if (GameInfo::getGameType() == GAME_TYPE_T7G)
		initShaders();
}

bool Win32PlatformInterface::IsMT32SupportedDevice()
{
	return true;
}

Common::String Win32PlatformInterface::getDefaultGraphicSetting()
{
	return SETTING_VALUE_GRAPHIC_MODE_HQ;
}

Common::String Win32PlatformInterface::getGameFilePath()
{
	switch (GameInfo::getGameType())
	{
	case GAME_TYPE_SIMON1:
		return "Simon1_data_all.bundle";
	case GAME_TYPE_SIMON2:
		return "Simon2_data_all.bundle";
	case GAME_TYPE_FOTAQ:
		return "FOTAQ_data_all.bundle";
	case GAME_TYPE_T7G:
		return "T7G_data_all.bundle";
	case GAME_TYPE_IHNM:
		return "IHNM_data_all.bundle";
	default:
		return ".";
	}
}

void Win32PlatformInterface::initShaders()
{
/*	mAreaVertex = compileShader(s_areaVertex, GL_VERTEX_SHADER);
	mAreaFragment = compileShader(s_areaFragment, GL_FRAGMENT_SHADER);
	mAreaProgram = linkShaders(mAreaVertex, mAreaFragment);

	mAreaPositionAttribute = glGetAttribLocation(mAreaProgram, "vPosition");
	mAreaProjMatUniform = glGetUniformLocation(mAreaProgram, "projMat");
	mAreaAlphaFactorUniform = glGetUniformLocation(mAreaProgram, "alphaFactor");*/
}


GLuint Win32PlatformInterface::compileShader(const Common::String &src, GLenum type) {
	/*int size = src.size();
	const char * source = src.c_str();
	GLuint shader = glCreateShader(type);
	glShaderSource(shader, 1, &source, &size);
	glCompileShader(shader);

	int status;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
	if (!status) {
		char * buffer;
		int length;
		glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &length);
		buffer = new char[length];
		glGetShaderInfoLog(shader, length, NULL, buffer);
		if (type == GL_VERTEX_SHADER)
			LOGD("Vertex shader compilation failed:\n%s", buffer);
		else
			LOGD("Fragment shader compilation failed:\n%s", buffer);
	}
	return shader;*/

	return 0;
}

GLuint Win32PlatformInterface::linkShaders(GLuint vertex, GLuint fragment) {
	/*GLuint program = glCreateProgram();
	if (vertex)
		glAttachShader(program, vertex);
	if (fragment)
		glAttachShader(program, fragment);
	glLinkProgram(program);
	int status;
	glGetProgramiv(program, GL_LINK_STATUS, &status);
	if (!status) {
		char * buffer;
		int length;
		glGetProgramiv(program, GL_INFO_LOG_LENGTH, &length);
		buffer = new char[length];
		glGetProgramInfoLog(program, length, NULL, buffer);
		LOGD("Shader program link failed:\n%s", buffer);
	}
	return program;*/

	return 0;
}